#include "sha256.h"

/* Required for htonl/ntohl. These could be replaced by macros, but since gcc
 * uses efficient builtin versions of this (e.g. the bswap instruction on x86)
 * this seems like a good solution for now. */
#include <arpa/inet.h>

#define ROR32(x,n)	(((x) >> (n))  |  ((x) << (32-(n))))

static const uint32_t h_init[8] = {
	0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
	0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19 };

static const uint32_t k[64] = {
	0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
	0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
	0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
	0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
	0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
	0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
	0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
	0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
	0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
	0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
	0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
	0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
	0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
	0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
	0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
	0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2 };

/* Mix the 128-bit state, h, with a 64-byte data block */
static void mix(uint32_t h[8], const void *data) {
	uint32_t w[64];
	uint32_t s0, s1, t1, t2;
	uint32_t A, B, C, D, E, F, G, H;
	int i;

	memcpy(w, data, 64);
	for(i=0; i<16; i++) w[i] = ntohl(w[i]);

	for(i=16; i<64; i++) {
		s0 = ROR32(w[i-15], 7) ^ ROR32(w[i-15], 18) ^ (w[i-15] >> 3);
		s1 = ROR32(w[i-2], 17) ^ ROR32(w[i-2], 19) ^ (w[i-2] >> 10);
		w[i] = w[i-16] + s0 + w[i-7] + s1;
	}

	A = h[0]; B = h[1]; C = h[2]; D = h[3];
	E = h[4]; F = h[5]; G = h[6]; H = h[7];

	for(i=0; i<64; i++) {
		s0 = ROR32(A, 2) ^ ROR32(A, 13) ^ ROR32(A, 22);
		t2 = s0 + ((A & B) ^ (A & C) ^ (B & C));
		s1 = ROR32(E, 6) ^ ROR32(E, 11) ^ ROR32(E, 25);
		t1 = ((E & F) ^ ((~E) & G)) + H + s1 + k[i] + w[i];
		H = G;
		G = F;
		F = E;
		E = D + t1;
		D = C;
		C = B;
		B = A;
		A = t1 + t2;
	}

	h[0] += A; h[1] += B; h[2] += C; h[3] += D;
	h[4] += E; h[5] += F; h[6] += G; h[7] += H;
}

/* Compute the SHA-256 hash of a data block in memory
 *
 * 		hash	32-byte buffer to store the hash sum
 * 		data	Data block to hash
 * 		len		Size of data block in bytes
 */
void sha256(void *hash, const void *data, size_t len) {
	uint32_t h[8];
	uint32_t t[16];
	uint64_t size;
	int i;

	size = ((uint64_t)len) * 8;
	memcpy(h, h_init, sizeof(h));
	while(len >= 64) {
		mix(h, data);
		data = ((char*)data) + 64;
		len -= 64;
	}

	memcpy(t, data, len);
	((uint8_t*)t)[len] = 0x80;
	if(len < 56) {
		memset(((uint8_t*)t)+len+1, 0, 56-(len+1));
		t[14] = htonl((size >> 32) & 0xffffffff);
		t[15] = htonl(size & 0xffffffff);
		mix(h, t);
		for(i=0; i<8; i++) h[i] = htonl(h[i]);
		memcpy(hash, h, sizeof(h));
		return;
	}
	memset(((uint8_t*)t)+len+1, 0, 64-(len+1));
	mix(h, t);
	memset(t, 0, 56);
	t[14] = htonl((size >> 32) & 0xffffffff);
	t[15] = htonl(size & 0xffffffff);
	mix(h, t);
	for(i=0; i<8; i++) h[i] = htonl(h[i]);
	memcpy(hash, h, sizeof(h));
}

