#include <stdio.h>
#include <string.h>
#include <time.h>

#include "aes.h"
#include "sha256.h"

static int test_sha(void) {
	char *str = "The quick brown fox jumps over the lazy dog";
	uint8_t correct[32] = {
		0xd7, 0xa8, 0xfb, 0xb3, 0x07, 0xd7, 0x80, 0x94,
		0x69, 0xca, 0x9a, 0xbc, 0xb0, 0x08, 0x2e, 0x4f,
		0x8d, 0x56, 0x51, 0xe4, 0x6d, 0x3c, 0xdb, 0x76,
		0x2d, 0x02, 0xd0, 0xbf, 0x37, 0xc9, 0xe5, 0x92
	};
	uint8_t hash[32];

	sha256(hash, str, strlen(str));
	return(memcmp(hash, correct, 32));
}

static int test_aes(void) {
#if Nk == 4
	uint8_t correct[16] = {
		0x69, 0xc4, 0xe0, 0xd8, 0x6a, 0x7b, 0x04, 0x30,
		0xd8, 0xcd, 0xb7, 0x80, 0x70, 0xb4, 0xc5, 0x5a
	};
#elif Nk == 6
	uint8_t correct[16] = {
		0xdd, 0xa9, 0x7c, 0xa4, 0x86, 0x4c, 0xdf, 0xe0,
		0x6e, 0xaf, 0x70, 0xa0, 0xec, 0x0d, 0x71, 0x91
	};
#elif Nk == 8
	uint8_t correct[16] = {
		0x8e, 0xa2, 0xb7, 0xca, 0x51, 0x67, 0x45, 0xbf,
		0xea, 0xfc, 0x49, 0x90, 0x4b, 0x49, 0x60, 0x89
	};
#endif
	uint8_t key[4*Nk];
	uint8_t plain[16], encrypted[16], decrypted[16];
	uint32_t w[Nrk], dw[Nrk];
	int i;

	for(i=0; i<4*Nk; i++) key[i] = i;
	for(i=0; i<16; i++) plain[i] = i*0x11;

	aes_expand_key(w, key);
	aes_encrypt_block(encrypted, plain, w);
	if(memcmp(encrypted, correct, 16)) return 1;
	aes_expand_dkey(dw, w);
	aes_decrypt_block(decrypted, encrypted, dw);
	if(memcmp(decrypted, plain, 16)) return 1;
	return 0;
}

static void xtea_encipher(int num_rounds, uint32_t *v, const uint32_t *k) {
	uint32_t v0, v1, sum=0, delta=0x9E3779B9;
	int i;
	v0 = v[0]; v1 = v[1];
	for(i=0; i<num_rounds; i++) {
		v0 += (((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum + k[sum & 3]);
		sum += delta;
		v1 += (((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum + k[(sum>>11) & 3]);
	}
	v[0] = v0; v[1] = v1;
}

static void xtea_decipher(int num_rounds, uint32_t *v, const uint32_t *k) {
	uint32_t v0, v1, delta=0x9E3779B9, sum=delta*num_rounds;
	int i;
	v0 = v[0]; v1 = v[1];
	for(i=0; i<num_rounds; i++) {
		v1 -= (((v0 << 4) ^ (v0 >> 5)) + v0) ^ (sum + k[(sum>>11) & 3]);
		sum -= delta;
		v0 -= (((v1 << 4) ^ (v1 >> 5)) + v1) ^ (sum + k[sum & 3]);
	}
	v[0] = v0; v[1] = v1;
}

static void test_speed(void) {
	uint8_t key[4*Nk];
	uint8_t buf[16];
	uint32_t w[Nrk], dw[Nrk];
	clock_t start, end;
	unsigned long i, n = 0x100000, m = 0x1000;
	uint8_t data[0x1000];
	uint8_t hash[32];
	double t, MB;

	for(i=0; i<4*Nk; i++) key[i] = i;
	for(i=0; i<16; i++) buf[i] = i*0x11;

	MB = ((double)n) / 65536.0;

	aes_expand_key(w, key);
	aes_expand_dkey(dw, w);

	start = clock();
	for(i=0; i<n; i++) {
		aes_encrypt_block(buf, buf, w);
	}
	end = clock();
	t = ((double)(end-start)) / (double)CLOCKS_PER_SEC;

	fprintf(stderr, "AES-%d encryption speed: %.02f MB/s\n", Nk*32, MB/t);

	start = clock();
	for(i=0; i<n; i++) {
		aes_decrypt_block(buf, buf, dw);
	}
	end = clock();
	t = ((double)(end-start)) / (double)CLOCKS_PER_SEC;

	fprintf(stderr, "AES-%d decryption speed: %.02f MB/s\n", Nk*32, MB/t);

	MB = ((double)m) / 256.0;
	memset(data, 0, sizeof(data));

	start = clock();
	for(i=0; i<m; i++) {
		sha256(hash, data, sizeof(data));
	}
	end = clock();
	t = ((double)(end-start)) / (double)CLOCKS_PER_SEC;

	fprintf(stderr, "SHA-256 speed: %.02f MB/s\n", MB/t);
}

static int test_xtea(void) {
	clock_t start, end;
	double t, MB;
	unsigned long i, n = 0x40000;
	uint32_t key[4] = { 0, 0, 0, 0 };
	uint32_t v[2] = { 0, 0 };
	uint32_t original[2];

	memcpy(original, v, sizeof(v));

	MB = ((double)n) / 131072.0;

	start = clock();
	for(i=0; i<n; i++) {
		xtea_encipher(64, v, key);
	}
	end = clock();
	t = ((double)(end-start)) / (double)CLOCKS_PER_SEC;

	fprintf(stderr, "XTEA encryption speed: %.02f MB/s\n", MB/t);

	start = clock();
	for(i=0; i<n; i++) {
		xtea_decipher(64, v, key);
	}
	end = clock();
	t = ((double)(end-start)) / (double)CLOCKS_PER_SEC;

	if(memcmp(v, original, 8)) return 1;

	fprintf(stderr, "XTEA decryption speed: %.02f MB/s\n", MB/t);

	return 0;
}

int main(void) {
	if(test_sha()) {
		fprintf(stderr, "SHA-256 test failed.\n");
		return 1;
	}

	if(test_aes()) {
		fprintf(stderr, "AES-%d test failed.\n", Nk*32);
		return 1;
	}

	test_speed();

	if(test_xtea()) {
		fprintf(stderr, "XTEA test failed.\n");
		return 1;
	}

	fprintf(stderr, "All tests passed.\n");
	return 0;
}

